/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001, 2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
 
   Original Author:Guruprasad Krishnamurthy <guruprasad.krishnamurthy@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: btctl-sdp.c,v 1.20 2002/12/04 20:09:29 kds Exp $

   Fixes:
		Manel Guerrero Zapata <manel.guerrero-zapata@nokia.com>
		Dmitry Kasatkin		: search added, fixes
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>


#include <affix/sdp.h>
#include <affix/sdpclt.h>

#include "btctl.h"

SDPServerHandle srvHandle;

char UUID_str[MAX_LEN_UUID_STR];

int _openConnection(char *target)
{
	int 			status;
	INQUIRY_ITEM 		inqItem[10];
	struct sockaddr_affix	sa;

	status = SDPInit(0);
	if (status < 0)
		return status;
	sa.family = PF_AFFIX;
	sa.local = BDADDR_ANY;
	if (target == NULL) {
		int duration = 5;
		uint8_t count = 0;

		status = SDPGeneralInquiry(inqItem, 10, duration, &count);
		if (count != 0) {
			sa.bda = inqItem[0].bda;
			printf("Connecting to host %s ...\n", bda2str(&sa.bda));
			srvHandle = SDPOpenServerConnection(&sa);
		}
	} else {
		if (strcmp(target, "local") == 0) {
			printf("Opening local connection\n");
			srvHandle = SDPOpenLocalServerConnection();
		} else {
			status = get_bda(&sa.bda, target);
			if (status) {
				printf("Incorrect address given\n");
				return 1;
			}
			printf("Connecting to host %s ...\n", bda2str(&sa.bda));
			srvHandle = SDPOpenServerConnection(&sa);
		}
	}
	if (srvHandle < 0)
		return -1;

	return 0;
}

void _closeConnection(void)
{
	if (srvHandle)
		SDPCloseServerConnection(srvHandle);
}

static void PrintGroups(void * value, void * userData)
{
	UUID *pUUID= NULL;

	pUUID = (UUID *)value;
	UUID2strn(pUUID, UUID_str, MAX_LEN_UUID_STR);
	if (UUID2value(pUUID) == PublicBrowseGroup)
		printf("  \"PublicBrowseGroup\" (0x%s)\n", UUID_str);
	else
		printf("  0x%s\n", UUID_str);
}


static void PrintServiceClass(void * value, void * userData)
{
	UUID *pUUID= NULL;
	char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];

	pUUID = (UUID *)value;
	UUID2strn(pUUID, UUID_str, MAX_LEN_UUID_STR);
	ServiceClassUUID2strn(pUUID, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
	printf("  \"%s\" (0x%s)\n", ServiceClassUUID_str, UUID_str);
}


static void PrintProtocolDescriptor(void * value, void * userData)
{
	ProtocolDescriptor *pPD = NULL;
	char ProtocolUUID_str[MAX_LEN_PROTOCOL_UUID_STR];

	pPD = (ProtocolDescriptor *)value;
	// printf("  Protocol Descriptors:\n");
	UUID2strn(&(pPD->protocolId), UUID_str, MAX_LEN_UUID_STR);
	ProtocolUUID2strn(&(pPD->protocolId), ProtocolUUID_str, MAX_LEN_PROTOCOL_UUID_STR);
	printf("  \"%s\" (0x%s)\n", ProtocolUUID_str, UUID_str);
	if (pPD->portNumber) {
		//printf("    PortSize: %d\n", pPD->portNumberSize);
		printf("    Port/Channel: %d\n", pPD->portNumber);
	}
	if (pPD->versionNumber) {
		printf("    Version: %#.4x\n", pPD->versionNumber);
	}
}


static void PrintAccessProtocols(void * value, void * userData)
{
	slist_t *protDescSeq = NULL;

	protDescSeq = (slist_t *)value;
	//printf("Access Protocols:\n");
	s_list_foreach(protDescSeq, PrintProtocolDescriptor, NULL);
}


static void PrintProfileDescriptor(void * value, void * userData)
{
	ProfileDescriptor *profDesc = NULL;
	char ProfileDescriptorUUID_str[MAX_LEN_PROFILEDESCRIPTOR_UUID_STR];

	profDesc = (ProfileDescriptor *)value;
	UUID2strn(&(profDesc->profileUUID), UUID_str, MAX_LEN_UUID_STR);
	ProfileDescriptorUUID2strn(&(profDesc->profileUUID), ProfileDescriptorUUID_str, MAX_LEN_PROFILEDESCRIPTOR_UUID_STR);
	printf("  \"%s\" (0x%s)\n", ProfileDescriptorUUID_str, UUID_str);
	if (profDesc->vnum) {
		printf("    Version: %#.4x\n", profDesc->vnum);
	}
}

int extractGroups(UUID *groupToExtract)
{
	int status = -1, i;
	uint16_t count = 0;
	slist_t *searchList = NULL;
	slist_t *attrIdList = NULL;
	slist_t *svcRecList = NULL;
	uint32_t range = 0x0000ffff;
	// uint32_t range = 0x00000004;
	// uint32_t range = 0x00000100;
	int groupIDStatus = -1;
	// UUID serviceClassUUID;
	slist_t *pslist_t = NULL;
	AccessProtocols *pAccessProtocols = NULL;

	s_list_append(&searchList, groupToExtract);
	s_list_append(&attrIdList, &range);
	status = SDPServiceSearchAttributeRequest(srvHandle, searchList, 
			RangeOfAttributes, attrIdList, 0xFFFF, &svcRecList, &count);
	//printf("Status : %d Count : %d\n", status, count);

	if (status || count == 0)
		return status;

	for (i = 0; i < (int)s_list_length(svcRecList); i++) {
		ServiceRecord *svcRec;
		printf("==============================\n");

		svcRec = (ServiceRecord *)s_list_nth_data(svcRecList, i);
		printSDPDataAsServiceRecord(svcRec);
		printf("------------------------------\n");

		printf("SvcRecHdl: 0x%x\n", svcRec->serviceRecordHandle);

		/*
		   printf("Service ID: \n");
		   getServiceID(bdAddr, svcRecHandle, &serviceClassUUID);
		   UUID2strn(&serviceClassUUID, UUID_str, MAX_LEN_UUID_STR);
		   printf("  0x%s\n", UUID_str);
		   PrintServiceClass(&serviceClassUUID, NULL);
		   */

		if (getServiceClassIDList(svcRec, &pslist_t) == E_OK) {
			printf("Service Class ID List: \n");
			s_list_foreach(pslist_t, PrintServiceClass, NULL);
		}

		if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
			printf("Protocol Descriptor List: \n");
			s_list_foreach(pAccessProtocols->pSequence, PrintAccessProtocols, NULL);
		}


		if (getProfileDescriptor(svcRec, &pslist_t) == E_OK) {
			printf("Profile Descriptor List: \n");
			s_list_foreach(pslist_t, PrintProfileDescriptor, NULL);
		}

		if (getBrowseGroupList(svcRec, &pslist_t) == E_OK) {
			printf("Browse Group List: \n");
			s_list_foreach(pslist_t, PrintGroups, NULL);
		}

		if (isGroup(svcRec)) {
			UUID groupId;

			// printf("SvcRec : 0x%x is a group\n", svcRecHandle);
			printf("This is a group.\n");
			groupIDStatus = getGroupID(svcRec, &groupId);
			// printf("Get Group ID status is %d\n", groupIDStatus);
			UUID2strn(&groupId, UUID_str, MAX_LEN_UUID_STR);
			printf("  0x%s\n", UUID_str);
			if (groupId.value.uuid16Bit != groupToExtract->value.uuid16Bit) {
				printf("Extracting it\n");
				extractGroups(&groupId);
			}
		}
	}
	
	return status;
}

int svcSearch(slist_t *searchList)
{
	int status = -1, i;
	uint16_t count = 0;
	slist_t *pSeq = NULL;
	slist_t *attrIdList = NULL;
	uint32_t range = 0x0000ffff;
	// uint32_t range = 0x00000004;
	// uint32_t range = 0x00000100;
	int groupIDStatus = -1;

	// UUID serviceClassUUID;
	slist_t *pslist_t = NULL;
	AccessProtocols *pAccessProtocols = NULL;

	// printf("In extractGroups method\n");
	s_list_append(&attrIdList, &range);
	status = SDPServiceSearchAttributeRequest(srvHandle, searchList, 
			RangeOfAttributes, attrIdList, 0xFFFF, &pSeq, &count);
	//printf("Status : %d Count : %d\n", status, count);

	if (status || count == 0)
		return status;

	for (i = 0; i < (int)s_list_length(pSeq); i++) {
		ServiceRecord	*svcRec;
		printf("==============================\n");

		svcRec = (ServiceRecord *)s_list_nth_data(pSeq, i);
		printSDPDataAsServiceRecord(svcRec);
		printf("------------------------------\n");

		printf("SvcRecHdl: 0x%x\n", svcRec->serviceRecordHandle);

		if (getServiceClassIDList(svcRec, &pslist_t) == E_OK) {
			printf("Service Class ID List: \n");
			s_list_foreach(pslist_t, PrintServiceClass, NULL);
		}

		if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
			printf("Protocol Descriptor List: \n");
			s_list_foreach(pAccessProtocols->pSequence, PrintAccessProtocols, NULL);
		}

		if (getProfileDescriptor(svcRec, &pslist_t) == E_OK) {
			printf("Profile Descriptor List: \n");
			s_list_foreach(pslist_t, PrintProfileDescriptor, NULL);
		}

		if (isGroup(svcRec)) {
			UUID groupId;

			printf("This is a group.\n");
			groupIDStatus = getGroupID(svcRec, &groupId);
			// printf("Get Group ID status is %d\n", groupIDStatus);
			UUID2strn(&groupId, UUID_str, MAX_LEN_UUID_STR);
			printf("  0x%s\n", UUID_str);
		}
	}

	return status;
}


int cmd_browse(struct btctl_command *cmd)
{
	int	status = -1;
	UUID 	uuid;

	status = _openConnection(argv[argind]);
	if (status != -1) {
		makeUUIDFrom16Bits(&uuid, PublicBrowseGroup);
		status = extractGroups(&uuid);
		_closeConnection();

	} else {
		printf("Connection to SDP server failed.\n");
	}
	return status;
}

int cmd_search(struct btctl_command *cmd)
{
	int		status = -1, i;
	slist_t		*svcSearchList = NULL;
	UUID 		uuid;
	UUID16Bit	searchArray[] = {
		LANAccessUsingPPPServiceClassID,
		DialupNetworkingServiceClassID,
		ObexObjectPushServiceClassID,
		ObexFileTransferServiceClassID,
		FaxServiceClassID,
		SerialPortServiceClassID,
		IrMCSyncServiceClassID,
		IrMCSyncCommandServiceClassID,
		HeadsetServiceClassID,
		HeadsetAudioGatewayServiceClassID,
		AGHandsFreeServiceClassID,
		HandsFreeServiceClassID
	};

	status = _openConnection(argv[argind]);
	if (status == -1) {
		printf("Connection to SDP server failed.\n");
		return status;
	}
	
	for (i = 0; i < (sizeof(searchArray)/sizeof(searchArray[0])); i++) {
		svcSearchList = NULL;
		makeUUIDFrom16Bits(&uuid, searchArray[i]);
		s_list_append(&svcSearchList, &uuid);
		status = svcSearch(svcSearchList);
		s_list_free(&svcSearchList);
	}
	_closeConnection();
	return status;
}


// FIXME: Remove LISTS!!!
