/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Based on the from Pontus Fuchs <pontus.fuchs@tactel.se>
   Copyright (c) 1999, 2000 Pontus Fuchs, All Rights Reserved.
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
   $Id: obex_socket.c,v 1.9 2002/11/01 15:13:30 kds Exp $

   OBEX serial port transport functions.

   Adapted for Affix, Fixes: 
   				Imre Deak <ext-imre.deak@nokia.com>
   				Dmitry Kasatkin
*/

#include <affix/config.h>

#include <stdio.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>

#include <stdint.h>

#include <openobex/obex.h>

#include <affix/btcore.h>
#include <affix/obex.h>


/* private */
int btobex_connect(obex_t *handle, void * userdata);
int btobex_listen(obex_t *handle, void * userdata);
int btobex_disconnect(obex_t *handle, void * userdata);
int btobex_write(obex_t *handle, void * userdata, uint8_t *buffer, int length);
int btobex_handle_input(obex_t *handle, void * userdata, int timeout);


obex_t *btobex_init(obex_link_t *link, obex_event_t eventcb, unsigned int flags)
{
	memset(link, 0, sizeof(obex_link_t));
	link->custfunc.userdata = link;
	link->rfd = link->rfd = -1;

	if (!(link->handle = OBEX_Init(OBEX_TRANS_CUST, eventcb, flags)))
		return NULL;

	link->custfunc.connect = btobex_connect;
	link->custfunc.disconnect = btobex_disconnect;
	link->custfunc.write = btobex_write;
	link->custfunc.handleinput = btobex_handle_input;
	link->custfunc.listen = btobex_listen;

	if (OBEX_RegisterCTransport(link->handle, &link->custfunc) < 0) {
		OBEX_Cleanup(link->handle);
		return NULL;
	}

	return link->handle;
}

//
// Close down. If force is TRUE. Try to break out of OBEX-mode.
//
void btobex_cleanup(obex_link_t *link, int force)
{
	BTDEBUG("");
	OBEX_Cleanup(link->handle);
}

//
// Open up cable OBEX
//
int btobex_setport(obex_link_t *link, const char *port)
{
	BTDEBUG("");
	link->rfd = link->wfd = -1;
	link->portname = port;
	return 0;
}

int btobex_setsocket(obex_link_t *link, const int fd)
{
	BTDEBUG("");
	link->rfd = link->wfd = fd;
	link->portname = NULL;
	return 0;
}

//
// Do transport connect
//
int btobex_connect(obex_t *handle, void * userdata)
{
	obex_link_t	*link = userdata;

	BTDEBUG("");
	
	if(link->rfd >= 0) {
		BTDEBUG("fd already exist. Using it");
		return 1;

	}
	if (link->rfd == -1) {
		if (link->portname) {
			// open device
			if ((link->rfd = open(link->portname, O_RDWR | O_NONBLOCK | O_NOCTTY, 0)) < 0) {
				perror("Can' t open tty");
				return -1;
			}
			link->wfd = link->rfd;

			tcgetattr(link->rfd, &link->oldtio);
			bzero(&link->newtio, sizeof(struct termios));
			link->newtio.c_cflag = B115200 | CS8 | CREAD; // T.Koba for RFCOMM
			link->newtio.c_iflag = IGNPAR;
			link->newtio.c_oflag = 0;
			tcflush(link->rfd, TCIFLUSH);
			tcsetattr(link->rfd, TCSANOW, &link->newtio);
		} else {
			link->rfd = 0;	// stdin
			link->wfd = 1;	// stdout
		}
	}

	return 1;
}

//
// Do transport listen
//
int btobex_listen(obex_t *handle, void * userdata)
{
	BTDEBUG("");
	
	return btobex_connect(handle, userdata);
}


//
// Do transport disconnect
//
int btobex_disconnect(obex_t *handle, void * userdata)
{
	obex_link_t	*link = userdata;

	BTDEBUG("");

	if (link->portname) {
		if (1/*force*/) {
			// Send a break to get out of OBEX-mode
			if (ioctl(link->rfd, TCSBRKP, 0) < 0) {
				BTERROR("Unable to send break!");
			}
		}
		close(link->rfd);
	} 
	link->rfd = link->wfd = -1;

	return 1;
}

//
//  Called when data needs to be written
//
int btobex_write(obex_t *handle, void * userdata, uint8_t *buffer, int length)
{
	obex_link_t	*link = userdata;
	int		actual = 0;
	int		pos = 0;

	while (actual < length)	{
		int frag;
		frag = write(link->wfd, &buffer[pos], length - pos);
		if (frag < 0) {
			if (errno == EAGAIN) {
				frag = 0;
				BTDEBUG("Temp. no resource avail.");
			} else {
				BTERROR("Write error:%s", strerror(errno));
				return frag;
			}
		}
		actual += frag;
		pos += frag;
		//BTDEBUG("Wrote %d fragment", frag);
	} 
	BTDEBUG("Wrote %d bytes (expected %d)", actual, length);
	return actual;
}

int btobex_handle_input(obex_t *handle, void * userdata, int timeout)
{
	int		ret, actual = 0;
	obex_link_t	*link = userdata;
	struct timeval	time;
	fd_set		fdset;
	
	BTDEBUG("reading from input...\n");

	if(link->rfd < 0)
		return -1;

	time.tv_sec = timeout;
	time.tv_usec = 0;

	FD_ZERO(&fdset);
	FD_SET(link->rfd, &fdset);

	ret = select(link->rfd+1, &fdset, NULL, NULL, &time);

	/* Check if this is a timeout (0) or error (-1) */
	if (ret < 1) {
		/* error */
		//BTDEBUG("ret (%d)", ret);
		return ret;
	}
	
	/* read data */
	actual = read(link->rfd, link->inputbuf, sizeof(link->inputbuf));
	if (actual < 0) {
		//BTDEBUG("read error (%d)", actual);
		return actual;
	}

	BTDEBUG("Read %d bytes", actual);

	/* pass up even if actual == 0 - openobex handle it well */
	return OBEX_CustomDataFeed(handle, link->inputbuf, actual);
}

